/*
** debug.c for elfsh
**
** All functions dealing with libedfmt uni debugging format
**
** Started on Fev 25 2007 mxatone
**
** $Id: debug.c,v 1.2 2007/03/07 16:45:35 thor Exp $
**
*/
#include "revm.h"

char buf[BUFSIZ];

static char 	*vm_debug_typename(u_int type)
{
  char 			*str;
  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  switch(type)
    {
    case EDFMT_TYPE_UNK:
      str = "?";
      break;
    case EDFMT_TYPE_BASIC:
      str = "basic";
      break;
    case EDFMT_TYPE_ARRAY:
      str = "array";
      break;
    case EDFMT_TYPE_PTR:
      str = "ptr";
      break;
    case EDFMT_TYPE_STRUCT:
      str = "struct";
      break;
    case EDFMT_TYPE_UNION:
      str = "union";
      break;
    case EDFMT_TYPE_ATTR:
      str = "attr";
      break;
    case EDFMT_TYPE_VOID:
      str = "void";
      break;
    case EDFMT_TYPE_LINK:
      str = "link";
      break;
    default:
      str = "?????";
    }

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, str);
}

static int		vm_debug_print_var(edfmtvar_t *vars)
{
  edfmtvar_t		*var;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  for (var = vars; var != NULL; var = var->next)
    {
      snprintf(buf, BUFSIZ - 1, " [%02d] %s: %08x => %s\n",
	       var->scope, var->name, var->addr, var->type->name);
      vm_output(buf);
    }

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

static int		vm_debug_print_type(edfmttype_t *types, u_int tab)
{
  edfmttype_t		*type;
  u_int			index;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  for (type = types; type != NULL; type = type->next)
    {
      for (index = 0; index < tab; index++)
	vm_output("\t");

      snprintf(buf, BUFSIZ - 1,
	       "%s - %s %d:%d ", vm_debug_typename(type->type), 
	       type->name, type->start, type->size);
      vm_output(buf);

      if (type->type != EDFMT_TYPE_ATTR && type->parent)
	{
	  snprintf(buf, BUFSIZ - 1,
		   "parent = 0x%p:%s ", type->parent, type->parent->name);
	  vm_output(buf);
	}

      if (type->type == EDFMT_TYPE_ATTR && type->child)
	{
	  snprintf(buf, BUFSIZ - 1,
		   "child = 0%p:%s ", type->child, type->child->name);
	  vm_output(buf);
	}

      vm_output("\n");
      if (type->type != EDFMT_TYPE_ATTR && type->child)
	{
	  vm_output(" => Childs:\n");
	  vm_debug_print_type(type->child, tab+1);
	}
    }

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

static int		vm_debug_unit_print_file(edfmtfile_t *files)
{
  edfmtfile_t		*file;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  for (file = files; file != NULL; file = file->next)
    {
      snprintf(buf, BUFSIZ - 1,
	       " ==== %s ===== %s\n", file->name,
	       file->parent ? file->parent->name : "-");
      vm_output(buf);

      vm_output(" .: Types :. \n");
      vm_debug_print_type(file->types, 0);
      vm_output("\n .: Vars :. \n");
      vm_debug_print_var(file->vars);

      if (file->child)
	vm_debug_unit_print_file(file->child);

      vm_output(" ============\n");
    }

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

int			cmd_debug()
{
  edfmtinfo_t		*info;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);
  
  info = edfmt_get_uniinfo(world.curjob->current);

  if (!info)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "No informations on this file", -1);

  vm_output(" .: Global Types :. \n");
  vm_debug_print_type(info->types, 0);
  vm_output("\n .: Global Vars :. \n");
  vm_debug_print_var(info->vars);

  vm_debug_unit_print_file(info->files);

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}
