/*
** inform.c for librevm in elfsh
**
** Inform a given type about a new variable
**
** Started Jan 21 2007 12:57:03 mayhem
**
**
** $Id: inform.c,v 1.11 2007/03/07 16:45:35 thor Exp $
**
*/
#include "revm.h"


/* Print registered variables for a given type */
int		vm_informed_print(char *name)
{
  hash_t	*hash;
  char		buf[BUFSIZ];
  revmvar_t	*tvar;
  char		**keys;
  int		index;
  int		keynbr;

  /* Preliminary checks */
  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);
  snprintf(buf, sizeof(buf), "type_%s", name);
  hash = hash_find(buf);
  if (!hash)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		      "Unknown informed type", -1);
  
  /* Retreive hash table content and print it */
  keys = hash_get_keys(hash, &keynbr);  
  if (keynbr)
    vm_output("  .:: Registered variables for this type \n\n");
  for (index = 0; index < keynbr; index++)
    {
      tvar = (revmvar_t *) hash_get(hash, keys[index]);
      snprintf(buf, sizeof(buf), "  + ["AFMT"] %-30s \n", tvar->addr, keys[index]);
      vm_output(buf);
    }

  /* End message */
  if (!keynbr)
    vm_output("  .:: No registered variables for this type \n\n");
  else
    {
      snprintf(buf, sizeof(buf), 
	       "\n .:: Displayed %u registered variables for type %s \n\n",
	       keynbr, name);
      vm_output(buf);
    }
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}




/* Add an element to the inform table for a given type */
int		vm_inform_type(char *type, char *varname, 
			       char *straddr, u_char print)
{
  hash_t	*hash;
  char		buf[BUFSIZ];
  elfsh_Addr	addr;
  elfsh_Addr	oaddr;
  char		*realname;
  char		*symname;
  revmvar_t	*tvar;
  elfsh_SAddr	off;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  /* Add element in the good hash table */
  snprintf(buf, sizeof(buf), "type_%s", type);
  hash = hash_find(buf);
  if (!hash)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		      "Unknown requested type", -1);
  
  /* The address is not precised, to to find it from the varname */
  if (!straddr)
    straddr = varname;

  /* The address is given, lookup it */
  realname = vm_lookup_string(varname);
  oaddr    = vm_lookup_addr(straddr);
  if (!oaddr)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		      "Invalid variable address", -1);
  if (!realname)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		      "Invalid variable name", -1);

  /* If just the address was given, lookup or create a name for the variable */
  if (IS_VADDR(realname))
    {
      sscanf(realname + 2, AFMT, &addr);
      symname = elfsh_reverse_symbol(world.curjob->current, addr, &off);
      if (symname && !off)
	realname = strdup(symname);
      else
	{
	  XALLOC(__FILE__, __FUNCTION__, __LINE__,realname, strlen(type) + 20, -1);
	  snprintf(realname, sizeof(realname), "%s_"AFMT, type, addr);
	}
    }

  addr = oaddr;
  XALLOC(__FILE__, __FUNCTION__, __LINE__, tvar, sizeof(revmvar_t), -1);

  /* Setup current string table offset */
  tvar->nameoff = vm_strtable_add(realname);

  /* Type will be set during save procedure */
  tvar->typenum = -1; 

  /* Default scope */
  tvar->scope = EDFMT_SCOPE_GLOBAL; 
  tvar->addr = addr;

  /* Overwrite existing entry */
  if (hash_add(hash, realname, (void*) tvar) < 0)
    {
      hash_del(hash, realname);
      if (hash_add(hash, realname, (void*) tvar) < 0)
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
			  "Add a new hash entry failed", -1);
    }

  if (print)
    {
      /* Success message and exit */
      snprintf(buf, sizeof(buf), 
	       " [*] Type %s succesfully informed of variable %s [" XFMT "]\n\n", 
	       type, realname, tvar->addr);
      vm_output(buf);
    }
  
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}



/* Add an element to the inform table for a given type */
int		vm_uninform_type(char *type, char *varname, u_char print)
{
  hash_t	*hash;
  char		buf[BUFSIZ];
  char		*realname;
  
  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);
   
  /* Grab element in the good hash table */
  snprintf(buf, sizeof(buf), "type_%s", type);
  hash = hash_find(buf);
  if (!hash)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		      "Unknown requested type", -1);

  /* In case we have to remove everything at once */
  if (!varname)
    {
      hash_empty(buf);
      if (print)
	{
	  snprintf(buf, sizeof(buf), 
		   " [*] Hash information for type %s succesfully reset \n\n", type);
	  vm_output(buf);
	}
      PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
    }

  /* In case we just have to remove one variable */
  realname = vm_lookup_string(varname);
  if (!realname)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		      "Invalid variable name", -1);

  if (!hash_get(hash, realname))
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		      "Unknown typed variable", -1);

  if (print)
    {
      snprintf(buf, sizeof(buf), 
	       " [*] Type %s succesfully uninformed of object %s \n\n", 
	       type, realname);
      vm_output(buf);
    }
  hash_del(hash, realname);
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}





/* Inform a given type */
int		cmd_inform()
{
  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);
  switch (world.curjob->curcmd->argc)
    {
      /* inform type */
    case 1:
      if (vm_informed_print(world.curjob->curcmd->param[0]) < 0)
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
			  "Invalid requested type", -1);
      break;

      /* inform type name */
    case 2:
      if (vm_inform_type(world.curjob->curcmd->param[0],
			 world.curjob->curcmd->param[1], NULL, 1) < 0)
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
			  "Invalid variable information", -1);
      break;
      
      /* inform type name address */
    case 3:
      if (vm_inform_type(world.curjob->curcmd->param[0],
			 world.curjob->curcmd->param[1],
			 world.curjob->curcmd->param[2], 1) < 0)
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
			  "Invalid variable information", -1);
      break;
    default:
      PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
			"Invalid inform syntax", -1);
    }
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}




/* Uninform a given type */
int		cmd_uninform()
{
  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);
  switch (world.curjob->curcmd->argc)
    {
      /* uninform type */
    case 1:
      if (vm_uninform_type(world.curjob->curcmd->param[0], NULL, 1) < 0)
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
			  "Invalid type information", -1);
      break;

      /* uninform type name */
    case 2:
      if (vm_uninform_type(world.curjob->curcmd->param[0],
			   world.curjob->curcmd->param[1], 1) < 0)
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
			  "Invalid variable information", -1);
      break;
    default:
      PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
			"Invalid uninform syntax", -1);
    }
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}
