/*
** script.c for librevm in elfsh
** 
** Started on  Fri May 16 15:18:35 2005 mayhem
**
**
** $Id: script.c,v 1.5 2007/03/07 16:45:35 thor Exp $
**
*/
#include "revm.h"


/* Script used as command function */
int			cmd_script()
{
  int                   idx;
  char                  data[BUFSIZ];
  u_int                 sz = BUFSIZ - 1;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  if (world.scriptsdir == NULL)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		      "No scripts dir specified", -1);

  snprintf(data, BUFSIZ - 1, "source %s/%s.esh",  
	   world.scriptsdir, world.curjob->curcmd->name);

  for (idx = 0; world.curjob->curcmd->param[idx] != NULL; idx++)
    {
      strncat(data, " ", sz - 1);
      sz -= 1;
      if (sz < 2)
        break;
      strncat(data, world.curjob->curcmd->param[idx], sz - 1);
      sz -= strlen(world.curjob->curcmd->param[idx]);
      if (sz < 2)
        break;
    }
  
  vm_exec_str(data);

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}


/* Set then scriptsdir directory and add cmd corresponding to scripts */
int	vm_add_script_cmd(char *dir_name)
{
  DIR	*dir;
  struct dirent *dir_entry;
  char	cmd_name[BUFSIZ];
  int	len;
  char  *argv[2];
  char	str[BUFSIZ];
  hashent_t     *actual;
  int           index;
  int	cnt = 0;


  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  argv[0] = "#!elfsh";

  dir = opendir(dir_name);
  
  if (dir == NULL)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		      "Opendir failed", (-1));
  

  /* scripts dir already specified so cleanup command set */
  if (world.scriptsdir != NULL)
    {
      for (index = 0; index < cmd_hash.size; index++)
	for (actual = &cmd_hash.ent[index];
	     actual != NULL && actual->key != NULL;
	     actual = actual->next)
	  {
	    if (((revmcmd_t *) actual->data)->exec == cmd_script)
	      {
		XFREE(__FILE__, __FUNCTION__, __LINE__,actual->data);
		hash_del(&cmd_hash, actual->key);
	      }
	  }
    }

  
  while( (dir_entry = readdir(dir)))
    {
      if ((len = strlen(dir_entry->d_name)) > 4)
	{
	  if (!strncmp(".esh", &(dir_entry->d_name[len - 1 - 3]), 4))
	    {
	      snprintf(str, BUFSIZ - 1, "%s/%s", dir_name, dir_entry->d_name);
	      argv[1] = str;
	      if (!vm_testscript(2, argv))
		continue;
	      
	      snprintf(cmd_name, BUFSIZ - 1, "%s", dir_entry->d_name);
	      cmd_name[len - 1 - 3] = '\0';
	      vm_addcmd(strdup(cmd_name), (void *) cmd_script, 
 
			(void *) vm_getvarparams, 0, 
			"Synthetic macro command");
	      if (!world.state.vm_quiet)
		{
		  snprintf(cmd_name, BUFSIZ - 1, "\t\t+ %s added\n", 
			   dir_entry->d_name);
		  vm_output(cmd_name);
		}
	      cnt++;
	    }
	}
    }
      
  if (world.scriptsdir != NULL)
    {
      XFREE(__FILE__, __FUNCTION__, __LINE__,world.scriptsdir);
      world.scriptsdir = NULL;
    }
  if (cnt)
    world.scriptsdir = strdup(dir_name);
 
  else 
    vm_output("\t\t [!] No scripts found \n");

  closedir(dir);
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}


/* Elfsh/E2dbg command for vm_add_script_cmd() */
int	cmd_scriptsdir()
{
  char	str[BUFSIZ];

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  if (world.curjob->curcmd->param[0] == NULL)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		      "Invalid parameter", (-1));
      
  if (!world.state.vm_quiet)
    {
      snprintf(str, BUFSIZ - 1, "\t [Adding script commands from %s] \n", 
	       world.curjob->curcmd->param[0]);
      vm_output(str);
    }
  
  if (vm_add_script_cmd(world.curjob->curcmd->param[0]) < 0)
    {
      vm_output("\t failed\n\n");
      PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
			"Macro command adding failed", (-1));
    }
  else
    {
#if defined(USE_READLN)
      readln_completion_commands(&cmd_hash);
#endif
      if (!world.state.vm_quiet)
	vm_output("\t done\n\n");
    }

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

int		cmd_lscripts()
{
  char		str[BUFSIZ];
  hashent_t     *actual;
  int           index;
  int		cnt = 1;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  /* scripts dir already specified so cleanup command set */
  if (world.scriptsdir != NULL)
    {
      vm_output(" .::. Script commands .::.\n");

      for (index = 0; index < cmd_hash.size; index++)
	for (actual = &cmd_hash.ent[index];
	     actual != NULL && actual->key != NULL;
	     actual = actual->next)
	  {
	    if (((revmcmd_t *) actual->data)->exec == cmd_script)
	      {
		snprintf(str, BUFSIZ - 1, " [%.3d] %s\n", cnt, actual->key);
		vm_output(str);
		cnt++;

	      }
	  }
      if (cnt == 1)
	vm_output("[*] No script commands\n");
      vm_output("\n");
    }  
  else
    vm_output(" [!!] No script directory specified, use sdir command.\n\n");

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}
