/*
** source.c for librevm in elfsh
** 
** Started on  Fri May 16 15:18:35 2005 mayhem
**
**
** $Id: source.c,v 1.12 2007/03/07 16:45:35 thor Exp $
**
*/
#include "revm.h"


/* Allow for the sourcing of elfsh scripts */
int		cmd_source()
{
  char		*str;
  int		fd;
  int		savedfd;
  char		savedmode;
  revmargv_t	*savedcmd;
  char          *(*savedinput)();  
  char		**argv;
  char		**av;
  int		ac;
  int           idx;
  char          actual[26];
  int		argc = 0 ;
  int		ret;
  revmobj_t	*new;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  /* build argc */
  for (idx = 0; world.curjob->curcmd->param[idx] != NULL; idx++)
    argc++;

  if (!idx)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
			  "Invalid parameter", (-1));

  /* Resolve the parameter */
  str = vm_lookup_string(world.curjob->curcmd->param[0]);
  if (!str)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "Invalid parameter", (-1));

  /* Switch to debugger script mode when sourcing */
  /* This does not include the debugger .elfshrc */
  if (!strstr(str, REVM_CONFIG) && world.state.vm_mode == REVM_STATE_DEBUGGER)
    world.state.vm_sourcing = 1;

  /* Open the script file and pass its parameters */
  argv = alloca((2 + argc) * sizeof(char *));
  argv[0] = "#!elfsh";
  argv[1] = str;
  ac = 1;
  for (ac = 1; world.curjob->curcmd->param[ac]; ac++)
    argv[ac + 1] = world.curjob->curcmd->param[ac]; 
  argv[ac + 1] = NULL;
  if (!vm_testscript(2, argv))
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "Invalid script file", -1);

  /* Add argument variables */
  for (idx = 1; argv[idx + 1]; idx++)
    {
      snprintf(actual, sizeof(actual), "%u", idx);
      new = vm_create_IMMEDSTR(1, strdup(argv[idx + 1]));
 
      hash_del(&vars_hash, actual);
      hash_add(&vars_hash, strdup(actual), new);
 
    }

  new = vm_create_IMMED(ASPECT_TYPE_INT, 1, idx);
  hash_del(&vars_hash, REVM_VAR_ARGC);
  hash_add(&vars_hash, REVM_VAR_ARGC, new);

  XOPEN(fd, str, O_RDONLY, 0, -1);

  world.curjob->sourced++;
  snprintf(actual, sizeof(actual), "job%u_labels", world.curjob->sourced);
  hash_init(&labels_hash[world.curjob->sourced], strdup(actual), 251, 
 
	    ASPECT_TYPE_STR);

  savedfd    = world.curjob->ws.io.input_fd;
  savedinput = world.curjob->ws.io.input;
  savedmode  = world.state.vm_mode;
  savedcmd   = world.curjob->curcmd;

  /* Temporary takes input from the script file */
  vm_setinput(&world.curjob->ws, fd);
  world.curjob->ws.io.input_fd = fd;
  world.state.vm_mode       = REVM_STATE_SCRIPT;
  world.curjob->ws.io.input    = vm_stdinput;

  /* A simple dedicated interpretation loop on the script file */
  do {

    av = vm_input(&ac);
    if (av == ((char **) REVM_INPUT_VOID))
      continue;
    else if (!av)
      break;

    if (vm_parseopt(ac, av) < 0)
      {
	XFREE(__FILE__, __FUNCTION__, __LINE__,av);
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
			  "Invalid script", (-1));
      }

    if (!world.curjob->curcmd                           || 
	(world.curjob->curcmd->name                     &&
	 (!strcmp(world.curjob->curcmd->name, CMD_QUIT) ||
	  !strcmp(world.curjob->curcmd->name, CMD_QUIT2))))
      {
	world.state.vm_sourcing = 0;
	break;
      }

  } while (1);
  
  /* We execute the parsed script and restore the original context */
  world.curjob->curcmd      = world.curjob->script[world.curjob->sourced];

  /* If we are executing a debugger script, do not clear everything yet */
  if (vm_execscript() == E2DBG_SCRIPT_CONTINUE)
    {
      world.context.savedfd    = savedfd;
      world.context.savedmode  = savedmode;
      world.context.savedcmd   = savedcmd;
      world.context.savedinput = savedinput;
      world.context.savedargv  = argv;
      world.context.savedname  = str;
      PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, E2DBG_SCRIPT_CONTINUE);
    }
  
  /* Restore the context immediately if not inside a debugger script */
  ret = vm_restore_dbgcontext(savedfd, 
			      savedmode, 
			      savedcmd, 
			      savedinput, 
			      argv, str);

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, ret);
}



/* Restore the previous debugger control context */
/* Mostly useful when using 'continue' command from a debugger script */
int	       vm_restore_dbgcontext(int		savedfd, 
				     char		savedmode, 
				     revmargv_t		*savedcmd, 
				     void		*savedinput, 
				     char		**argv,
				     char		*savedname)
{
  char		buf[BUFSIZ];
  u_int		idx;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  /* Restore modified control context */
  world.curjob->ws.io.input_fd = savedfd;
  world.state.vm_mode       = savedmode;
  world.curjob->curcmd      = savedcmd;
  world.curjob->ws.io.input    = savedinput;
  snprintf(buf, BUFSIZ, "\n [*] %s sourcing -OK- \n\n", savedname);
  vm_output(buf);
  
  world.curjob->script[world.curjob->sourced] = NULL;
  hash_destroy(&labels_hash[world.curjob->sourced]);
  world.curjob->sourced--;

  /* Del argument variables */
  for (idx = 1; argv[idx + 1]; idx++)
    {
      snprintf(buf, sizeof(buf), "%u", idx);
      hash_del(&vars_hash, buf);
    }
  hash_del(&vars_hash, REVM_VAR_ARGC);

  /* Return OK */
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}




/* Execute a command given by a string in the current job */
int	vm_exec_str(char *str)
{
  int	nbr;
  char  **av;
  int	ac;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);
  nbr = vm_findblanks(str);
  av = vm_doargv(nbr, (u_int *)&ac, str);
  if (vm_parseopt(ac, av) < 0)
    {
      XFREE(__FILE__, __FUNCTION__, __LINE__,av);
      PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
			 "Command failed", (-1));
    }
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}
