/*
** types.c for librevm in ERESI
** 
** Be able to add your own types in runtime
**
** Started on  Sun Jan 9 07:23:58 2007 jfv
**
** $Id: types.c,v 1.20 2007/03/07 16:45:35 thor Exp $
**
*/
#include "revm.h"

/* Copy a designed type */
int		vm_type_copy(char *from, char *to)
{
  aspectype_t	*tocopy, *newtype;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  tocopy = (aspectype_t *) hash_get(&types_hash, from);
  if (!tocopy)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		 "Type not found", -1);

  newtype = (aspectype_t *) hash_get(&types_hash, to);

  if (newtype)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		 "Type destination already exist", -1);    

  newtype = aspect_type_copy(tocopy, tocopy->off, 
			     tocopy->isptr, tocopy->dimnbr,
			     strdup(to), tocopy->elemnbr);
  if (!newtype)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		 "Failed to create a new type", -2);

  hash_add(&types_hash, newtype->fieldname, (void *) newtype);
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

/* Init the global type table */
int		vm_type_hashcreate(char *name)
{
  char		hashname[BUFSIZ];
  hash_t	*newhash;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);
  
  /* Create the hash table for objects of that type */
  snprintf(hashname, sizeof(hashname), "type_%s", 
	   name);
  XALLOC(__FILE__, __FUNCTION__, __LINE__, newhash, sizeof(hash_t), -1);
  hash_init(newhash, strdup(hashname), 11, ASPECT_TYPE_UNKNOW);

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

/* Print a specific type */
int		vm_type_print(char *type, char mode)
{
  aspectype_t	*cur;
  aspectype_t	*child;
  char		buf[BUFSIZ];
  char		prefix[128];
  int		len;
  char		*size;
  char		offset[128];
  int		idx;
  int		sz;
  char		*pad;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  cur = hash_get(&types_hash, type);
  if (!cur)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		      "Unknown type", 0);

  /* Set up things */
  vm_endline();

  /* Setup first part of the line */
  snprintf(prefix, sizeof(prefix), "%s%s%s", 
	   vm_colorfieldstr("{"),
	   vm_colornumber("%u", cur->size),
	   vm_colorfieldstr("}"));
  len = snprintf(buf, sizeof(buf), "  %s %-20s %s %-10s", 
		 vm_colorfieldstr("Type"),
		 vm_colortypestr_fmt("%-20s", type), 
		 vm_colorfieldstr("size"),
		 prefix);
  size = alloca(20);

  /* If the type is a structure */
  if (cur->childs)
    {
      len += snprintf(buf + len, sizeof(buf) - len, "%s",
		      vm_colorfieldstr(" = {"));
      
      /* Prepare the padding after each field name */
      sz = len - vm_color_size(buf) - 14; /* -14 is dirty: some bug in colors */
      pad = alloca(sz + 1);
      memset(pad, ' ', sz);
      pad[sz] = 0x00;

      /* For each child field type */
      for (child = cur->childs; child; child = child->next)
	{
	  
	  /* Compute the size field */
	  if (child->type == ASPECT_TYPE_RAW)
	    snprintf(size, sizeof(size), "%s%s%s", 
		     vm_colorfieldstr("["), 
		     vm_colornumber("%u", child->size), 
		     vm_colorfieldstr("]"));
	  else if (child->dimnbr)
	    {
	      for (sz = idx = 0; idx < child->dimnbr; idx++)
		sz += 20;
	      size = alloca(sz);
	      for (sz = idx = 0; idx < child->dimnbr; idx++)
		sz += sprintf(size + sz, "%s%s%s", 
			      vm_colorfieldstr("["),
			      vm_colornumber("%u", child->elemnbr[idx]),
			      vm_colorfieldstr("]"));
	    }
	  else
	    *size = 0x00;

	  /* Format the whole thing */
	  if (mode)
	    *offset = 0x00;
	  else
	    snprintf(offset, sizeof(offset), "%s%s", 
		     vm_colornumber("%u", child->off),
		     vm_colorfieldstr(":"));
	  len += snprintf(buf + len, sizeof(buf) - len, 
			  "%s%s%s%s%s%s%s", 
			  offset, 
			  vm_colorstr(child->fieldname),
			  vm_colorfieldstr(":"),
			  vm_colortypestr((child->isptr ? "*" : "")), 
			  vm_colortypestr(child->name), 
			  size,
			  vm_colorfieldstr((child->next ? ",\n" : "}\n\n")));
	  
	  /* Print field and next padding */
	  vm_output(buf);
	  vm_endline();
	  if (child->next)
	    vm_output(pad);
	  *buf = 0x00;
	  len = 0;
	}
    }

  /* Print non-structures types */
  else
    {
      vm_output(buf);
      vm_output("\n");
    }

  /* Return success */
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 1);
}


/* Print types matching a regular expression */
int		vm_type_print_regex(char *regex)
{
  regex_t	rx;
  int		keynbr;
  char		**keys;
  int		index;
  int		match;
  char		buf[50];

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  if (regcomp(&rx, regex, REG_EXTENDED) < 0)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		 "Failed to compute regex", -1);

  keys = hash_get_keys(&types_hash, &keynbr);
  for (match = index = 0; index < keynbr; index++)
    {
      if (!regexec(&rx, keys[index], 0, 0, 0))
	{
	  vm_type_print(keys[index], 0);
	  match++;
	}
    }

  snprintf(buf, sizeof(buf), " [*] Matched %u types \n\n", match);
  vm_output(buf);
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);  
}



/* Print available types */
int		vm_type_prints()
{
  int		keynbr;
  char		**keys;
  int		index;
  int		total;
  char		buf[BUFSIZ];

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);
  if (!types_hash.elmnbr)
    {
      vm_output("\n  .:: No registered types\n\n");
      PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
    }
  vm_output("\n  .:: Available language types \n\n");
  keys = hash_get_keys(&types_hash, &keynbr);
  for (total = index = 0; index < keynbr; index++)
    total += vm_type_print(keys[index], 1);
  snprintf(buf, sizeof(buf), "\n  .:: Found %u registered types\n\n", total);
  vm_output(buf);
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}


/* Types management command */
int		cmd_type()
{
  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  switch (world.curjob->curcmd->argc)
    {
      /* type */
    case 0:
      vm_type_prints();
      break;

      /* type name */
    case 1:
      vm_type_print_regex(world.curjob->curcmd->param[0]);
      vm_output("\n");
      break;

      /* Those case are invalid syntaxes */
    case 2:
      PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
			"Invalid syntax for type command", -1);

    default:
      /* Add a type */
      if (!strcmp(world.curjob->curcmd->param[1], "=") ||
	  !strcmp(world.curjob->curcmd->param[1], "is"))
	{
	  if (aspect_type_register(world.curjob->curcmd->param[0],
				   world.curjob->curcmd->param + 2,
				   world.curjob->curcmd->argc - 2) < 0)
	    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
			 "Invalid type declaration", -1);

	  vm_output("\n\t [*] Added type ");
	  vm_output(world.curjob->curcmd->param[0]);
	  vm_output(" succesfully \n\n");

	  /* Create the hash table for objects of that type */
	  vm_type_hashcreate(world.curjob->curcmd->param[0]);
	}
      else
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		     "Invalid syntax for type command", -1);
 
    }
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

/* Typedef command function */
int  		cmd_typedef()
{
  int		ret;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  if (world.curjob->curcmd->argc < 2)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		 "Invalid syntax for type command", -1);

  ret = vm_type_copy(world.curjob->curcmd->param[0], 
		     world.curjob->curcmd->param[1]);

  if (ret < 0)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		 (ret == -1 ? "Type not found" : 
		  "Failed to create a new type"), -1);
  
  vm_output("\n\t [*] Copied type ");
  vm_output(world.curjob->curcmd->param[0]);
  vm_output(" to ");
  vm_output(world.curjob->curcmd->param[1]);
  vm_output(" succesfully \n\n");

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}
